/*
 -- Multicore --
 fixed point test code:
timeing varies, but generally s15x16 is 3 to 4 times
faster than float and s1x14 is 10 to 12 times faster
Example:
 input iterations: 125
fptime 171  fp_product 1.734370
fix16time 50 fix_product 1.731339
fix14time 12 fix_product 1.716187

For divide,  s15x16 is about 1.4 times
faster than float and s1x14 is about 2.8 times faster
but note that s1x14 cannot represent any division with 
a quotient greater than 2.
200 iterations
fptime 165  fp_div 0.136687
fix16time 117 fix_div 0.136169
fix14time 59 fix_div 0.135376

For sqrt, s15x16 is about 1.4 times faster than float

 */

#include "hardware/gpio.h"
#include "hardware/timer.h"

#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "stdio.h"
#include <string.h>
#include <pico/multicore.h>
#include <pico/divider.h>
#include "hardware/sync.h"
#include "stdlib.h"
#include "math.h"

// ==========================================
// === fixed point s15x16
// ==========================================
// s15x16 fixed point macros ==
// == resolution 2^-16 = 1.5e-5
// == dynamic range is 32767/-32768
typedef signed int s15x16;
#define muls15x16(a,b) ((s15x16)(((( signed long long )(a))*(( signed long long )(b)))>>16)) //multiply two fixed 16:16
#define float_to_s15x16(a) ((s15x16)((a)*65536.0)) // 2^16
#define s15x16_to_float(a) ((float)(a)/65536.0)
#define s15x16_to_int(a)    ((int)((a)>>16))
#define int_to_s15x16(a)    ((s15x16)((a)<<16))
#define divs15x16(a,b) ((s15x16)((((signed long long)(a)<<16)/(b)))) 
#define abss15x16(a) abs(a)

// see below for better sqrt
//#define sqrts15x16(a) (float_to_s15x16(sqrt(s15x16_to_float(a)))) 
/*
Algorithm and code Author Christophe Meessen 1993. 
  Initially published in usenet comp.lang.c, Thu, 28 Jan 1993 08:35:23 GMT, 
  Subject: Fixed point sqrt ; by Meessen Christophe
  https://groups.google.com/forum/?hl=fr%05aacf5997b615c37&fromgroups#!topic/comp.lang.c/IpwKbw0MAxw/discussion
  Note: there was a bug in the published sqrtL2L routine. It is corrected in
  this implementation.
  This code from https://github.com/chmike/fpsqrt/blob/master/fpsqrt.c
*/
// the squrare root of a fixed point with 16 bit
// fractional part and returns a fixed point with 16 bit fractional part. It 
// requires that v is positive. The computation use only 32 bit registers and 
// simple operations.
s15x16 sqrts15x16(s15x16 v) {
    uint32_t t, q, b, r;
    r = v;
    b = 0x40000000;
    q = 0;
    while( b > 0x40 )
    {
        t = q + b;
        if( r >= t )
        {
            r -= t;
            q = t + b; // equivalent to q += 2*b
        }
        r <<= 1;
        b >>= 1;
    }
    q >>= 8;
    return q;
}

// ==========================================
// === fixed point s1x14
// ==========================================
// s1.14 format
// == resolution 2^-14 = 6.1035e-5
// == dynamic range is +1.9999/-2.0
typedef signed short s1x14 ;
#define muls1x14(a,b) ((s1x14)((((int)(a))*((int)(b)))>>14)) 
#define float_to_s1x14(a) ((s1x14)((a)*16384.0)) // 2^14
#define s1x14_to_float(a) ((float)(a)/16384.0)
#define abss1x14(a) abs(a) 
#define divs1x14(a,b) ((s1x14)((((signed int)(a)<<14)/(b)))) 

// ==========================================
// === protothreads globals
// ==========================================
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

// ==================================================
// === user's serial input thread
// ==================================================
// test speed of fixed point
static PT_THREAD (protothread_serial(struct pt *pt))
{
  PT_BEGIN(pt);
    static int num_iterations ;
    static int fptime, fixtime, fix14time ;
    //
    while(1) {
      // print prompt
      sprintf(pt_serial_out_buffer, "input iterations: ");
      // spawn a thread to do the non-blocking write
      serial_write ;

      // spawn a thread to do the non-blocking serial read
      serial_read ;
      // convert input string to number
      sscanf(pt_serial_in_buffer,"%d ", &num_iterations) ;

      //do fp mult and add
      float c1 = 0.01, fp1 = .5;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = fp1 + (fp1 * c1) ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do s15x16 mult and add
      s15x16 d1 = float_to_s15x16(0.01), fix1 = float_to_s15x16(.5);
      fixtime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix1 = fix1 + muls15x16(fix1, d1) ;
      }
      fixtime = PT_GET_TIME_usec() - fixtime ;

      //do s1x14 MAC
      s1x14 e1 = float_to_s1x14(0.01), fix14_1 = float_to_s1x14(.5);
      fix14time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix14_1 = fix14_1 + muls1x14(fix14_1, e1) ;
      }
      fix14time = PT_GET_TIME_usec() - fix14time ;
      
      // timeing varies, but generally s15x16 is 3 to 4 times
      // faster than floaqt and s1x14 is 10 to 12 times faster
      // print summary of timing
      printf("fptime %d  fp_mac %f \r\n", fptime, fp1) ;
      printf("fix16time %d fix_mac %f \r\n", fixtime, s15x16_to_float(fix1));
      printf("fix14time %d fix_mac %f \r\n", fix14time, s1x14_to_float(fix14_1));
      printf("\r\n") ;
      // =================
      //do fp divide
      c1 = 1.01 ; 
      fp1 = 1.0;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = fp1 / c1 ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do s15x16 divide
      d1 = float_to_s15x16(1.01);
       fix1 = float_to_s15x16(1.0);
      fixtime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix1 = divs15x16(fix1, d1) ;
      }
      fixtime = PT_GET_TIME_usec() - fixtime ;

      //do s1x14 
      e1 = float_to_s1x14(1.01);
      fix14_1 = float_to_s1x14(1.0);
      fix14time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix14_1  =  divs1x14(fix14_1, e1) ;
      }
      fix14time = PT_GET_TIME_usec() - fix14time ;
      
      // timeing varies, but generally s15x16 is 3 to 4 times
      // faster than floaqt and s1x14 is 10 to 12 times faster
      // print summary of timing
      printf("fptime %d  fp_div %f \r\n", fptime, fp1) ;
      printf("fix16time %d fix_div %f \r\n", fixtime, s15x16_to_float(fix1));
      printf("fix14time %d fix_div %f \r\n", fix14time, s1x14_to_float(fix14_1));
      printf("\r\n") ;

// =================
      //do fp sqrt
      c1 = 1.01 ; 
      fp1 = 1.0;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = fp1 + sqrt(fp1);
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do s15x16 sqrt
      d1 = float_to_s15x16(1.01);
      fix1 = float_to_s15x16(1.0);
      fixtime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix1 = fix1 + sqrts15x16(fix1) ;
      }
      fixtime = PT_GET_TIME_usec() - fixtime ;

      // timeing varies
      printf("fptime %d  fp_sqrt %f \r\n", fptime, fp1) ;
      printf("fix16time %d fix_sqrt %f \r\n", fixtime, s15x16_to_float(fix1));
      //printf("fix14time %d fix_div %f \r\n", fix14time, s1x14_to_float(fix14_1));
      printf("\r\n") ;


      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // timer thread

// ==================================================
// === toggle25 thread 
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
  PT_BEGIN(pt);
    static bool LED_state = false ;
    
    // set up LED p25 to blink
    gpio_init(25) ;	
    gpio_set_dir(25, GPIO_OUT) ;
    gpio_put(25, true);
    // data structure for interval timer
    PT_INTERVAL_INIT() ;
     
    while(1) {
      // yield time 0.1 second
      PT_YIELD_INTERVAL(100000) ;

      // toggle the LED on PICO
      LED_state = LED_state? false : true ;
      gpio_put(25, LED_state);
      //
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ==================================================
// === toggle_fast thread -- RUNNING on core 1
// ==================================================
// toggle gpio 4 
static PT_THREAD (protothread_toggle_fast(struct pt *pt))
{
  PT_BEGIN(pt);
    static bool LED_state = false ;
    //
    // set up LED gpio 4 to blink
    gpio_init(4) ;	
    gpio_set_dir(4, GPIO_OUT) ;
    gpio_put(4, true);
    // data structure for interval timer
    PT_INTERVAL_INIT() ;

    while(1) {
      //
      PT_YIELD_INTERVAL(20) ;
      // toggle gpio 4
      LED_state = !LED_state ;
      gpio_put(4, LED_state);
      //
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_toggle_fast);
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");
     
  // === add threads ======================
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ===================
  // for core 0
  pt_add_thread(protothread_serial);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main
///////////
// end ////
///////////